#include <hyperConfig.hh>

#include <fstream>
#include <sstream>

#include <compiler/ability.hh>
#include <compiler/expression_ast.hh>
#include <compiler/exec_expression_output.hh>
#include <compiler/extract_symbols.hh>
#include <compiler/functions_def.hh>
#include <compiler/logic_expression_output.hh>
#include <compiler/output.hh>
#include <compiler/scope.hh>
#include <compiler/task.hh>
#include <compiler/types.hh>
#include <compiler/universe.hh>

#include <genom_implementation.hh>

using namespace hyper::compiler;

void 
genom_extension::function_proto(
		std::ostream& oss, const functionDef& f, const typeList& tList) const
{
	oss << "\t\tstruct " << scope::get_identifier(f.name()) << " {\n";
	type ret = tList.get(f.returnType());
	oss << "\t\t\ttypedef " << ret.type_name() << " output_type;\n"; 
	oss << "\t\t\ttypedef void genom_input_type; // XXX to be redefined\n";
	oss << "\t\t\ttypedef void genom_output_type; // XXX to be redefined\n\n";

	oss << "\t\t\tstatic void callback(const boost::system::error_code& e, \n";
	oss << times(8, "\t") << "genom_model::reference<genom_output_type>::type genom_output, \n";
	oss << times(8, "\t") << "genom_model::reference<output_type>::type output, \n";
	oss << times(8, "\t") << "hyper::genom_model::cb_type cb);\n\n";

	oss << "\t\t\tstatic ssize_t apply (";
	for (size_t i = 0; i < f.arity(); ++i) 
	{
		type arg = tList.get(f.argsType(i));
		oss << arg.type_name();
		if (arg.t == stringType || arg.t == structType || arg.t == opaqueType)
			oss << " const &";
		else
			oss << " ";
		oss << ", ";
	}
	oss << "\n" << times(8, "\t") << "boost::asio::io_service&, \n";
	oss << times(8, "\t") << "genom_model::reference<genom_input_type>::type genom_input, \n";
	oss << times(8, "\t") << "genom_model::reference<genom_output_type>::type genom_output, \n";
	oss << times(8, "\t") << "genom_model::reference<output_type>::type output, \n";
	oss << times(8, "\t") << "hyper::genom_model::cb_type);\n";

	// construct abort only if first element is the "opaque type representing genom module"
	// Of course, the test here is pretty buggy ...
	if (f.arity() > 0) {
		type arg = tList.get(f.argsType(0));
		if (arg.t == opaqueType) { 
			oss << "\n\t\t\tstatic void abort (const " << arg.type_name() << "& v0, ssize_t v1) {\n";
			oss << "\t\t\t\t return v0.impl->abort(v1); \n";
			oss << "\t\t\t}\n";
		} 
	}

	oss << "\t\t};\n\n";
	oss << std::endl;
}

void genom_extension::function_proto_additional_includes(
		std::ostream& oss, const std::string& name) const
{
	oss << "#include <genom_model/reference.hh>\n";
	oss << "#include <" << name << "/" << name << "_genom_interface.hh>\n\n"; 
}

void genom_extension::function_impl(
		std::ostream& oss, const functionDef& f, const typeList& tList) const
{
	oss << "\t\tvoid " << scope::get_identifier(f.name()) << "::callback";
	oss << "(const boost::system::error_code& e, \n";
	oss << times(6, "\t") << "genom_model::reference<genom_output_type>::type genom_output, \n";
	oss << times(6, "\t") << "genom_model::reference<output_type>::type output, \n";
	oss << times(6, "\t") << "hyper::genom_model::cb_type cb)\n\n";
	oss << times(2, "\t") << "{\n";
	oss << times(3, "\t") << "if (e)\n";
	oss << times(4, "\t") << "return cb(e);\n\n";
	oss << times(3, "\t") << "// convert genom_output to ouput if necessary\n\n";
	oss << times(3, "\t") << "cb(e);\n";
	oss << times(2, "\t") << "}\n\n";

	type ret = tList.get(f.returnType());
	oss << "\t\tssize_t " << scope::get_identifier(f.name()) << "::apply(";
	for (size_t i = 0; i < f.arity(); ++i) 
	{
		type arg = tList.get(f.argsType(i));
		oss << arg.type_name();
		if (arg.t == stringType || arg.t == structType || arg.t == opaqueType)
			oss << " const & ";
		else
			oss << " ";
		oss << "v" << i;
		oss << ", ";
	}

	oss << "\n" << times(6, "\t") << "boost::asio::io_service& io_s, \n";
	oss << times(6, "\t") << "genom_model::reference<genom_input_type>::type genom_input, \n";
	oss << times(6, "\t") << "genom_model::reference<genom_output_type>::type genom_output, \n";
	oss << times(6, "\t") << "genom_model::reference<output_type>::type output, \n";
	oss << times(6, "\t") << "hyper::genom_model::cb_type cb)\n";
	oss << "\t\t{\n"; // #error " << scope::get_identifier(f.name()) << " not implemented !!\n";
	oss << times(3, "\t") << "hyper::genom_model::cb_type local_cb = \n";
	oss << times(5, "\t") << "boost::bind(" << scope::get_identifier(f.name()) << "::callback, \n";
	oss << times(6, "\t") << "boost::asio::placeholders::error,\n";
	oss << times(6, "\t") << "boost::ref(genom_output), boost::ref(output), cb);\n";
	oss << "\t\t}" << std::endl;
}

struct extract_fun_name : boost::static_visitor<std::string>
{
	template <typename T>
	std::string operator() (const T&) const { assert(false); return ""; }

	std::string operator() (const function_call& f) const { return f.fName; }
};

struct genom_expression_ast_output_helper : boost::static_visitor<std::string>
{
	const std::set<std::string>& remote_syms;
	const symbolList& syms;

	genom_expression_ast_output_helper(const std::set<std::string>& remote_syms,
									   const symbolList& syms) :
		remote_syms(remote_syms), syms(syms)
	{}

	template <typename T>
	std::string operator() (const T&) const { assert(false); return ""; }

	std::string operator() (const function_call& f) const { 
		std::ostringstream oss;
		oss << f.fName << "::apply(";
		for (size_t i = 0; i < f.args.size(); ++i) {
			oss << expression_ast_output(f.args[i], remote_syms, syms);
			oss << ", ";
		}
		oss << "a.io_s, genom_input, genom_output, output, cb)";
		return oss.str();
	}
};


std::string genom_expression_ast_output(const expression_ast& e, const std::set<std::string>& remotes, 
			   						    const symbolList& local_symbol)
{
	return boost::apply_visitor(genom_expression_ast_output_helper(remotes, local_symbol), e.expr);
}

struct abort_visitor : boost::static_visitor<std::string>
{
	const ability& a;
	const universe& u;
	const std::set<std::string>& remote_syms;
	const symbolList& syms;

	abort_visitor(const ability& a, const universe& u, 
				  const std::set<std::string>& remote_syms,
				  const symbolList& syms) : 
		a(a), u(u), remote_syms(remote_syms), syms(syms) 
	{}

	template <typename T>
	std::string operator() (const T&) const { assert(false); return ""; }

	std::string operator() (const function_call& f) const {
		std::string full_name = scope::add_scope(a.name(), f.fName);
		std::pair<bool, functionDef> p = u.funcs().get(full_name);
		assert(p.first);
		if (p.second.arity() > 0) {
			type t = u.types().get(p.second.argsType(0));
			if (t.t != opaqueType) 
				return "";

			std::ostringstream oss;
			oss << f.fName << "::abort(";
			oss << expression_ast_output(f.args[0], remote_syms, syms);
			oss << ", id)";
			return oss.str();
		} 

		return "";
	}
};

void genom_extension::dump_eval_expression(std::ostream& oss, const universe& u,
		const ability& a, const task& t,
		const std::string& local_data_type, const symbolList& local_symbol,
		size_t counter, const expression_ast& e) const
{
	std::string indent = "\t\t";
	std::string next_indent = "\t" + indent;
	oss << indent << "struct genom_expression_" << counter++ << " {\n";

	/* Compute the necessary arguments */
	extract_symbols syms(a);
	syms.extract(e);
	oss << next_indent << "typedef model::update_variables< ";
	oss << "hyper::" << a.name() << "::ability, ";
	oss << syms.local_with_updater.size() << ", ";
	oss << syms.remote_vector_type_output(u) << " > updater_type;\n";

	/* Extract name of function call */
	std::string function_name = boost::apply_visitor(extract_fun_name(), e.expr);
	oss << next_indent << "typedef " << function_name << "::genom_input_type genom_input_type;\n";
	oss << next_indent << "typedef " << function_name << "::genom_output_type genom_output_type;\n";
	oss << next_indent << "typedef " << function_name << "::output_type output_type;\n";

	oss << next_indent << "ssize_t operator() (const updater_type& updater, ";
	oss << "hyper::" << a.name() << "::ability& a,\n";
	oss << times(9, "\t") << "const " << local_data_type << " & local_vars,\n";
	oss << times(9, "\t") << "genom_model::reference<genom_input_type>::type genom_input, \n";
	oss << times(9, "\t") << "genom_model::reference<genom_output_type>::type genom_output, \n";
	oss << times(9, "\t") << "genom_model::reference<output_type>::type output, \n";
	oss << times(9, "\t") << "hyper::genom_model::cb_type cb)\n";
	oss << next_indent << "{\n";
	oss << next_indent << "\treturn " << genom_expression_ast_output(e, syms.remote, local_symbol) << ";\n";
	oss << next_indent << "}\n";

	oss << next_indent << "static void abort(hyper::" << a.name() << "::ability& a, ";
	oss << "const " << local_data_type << "&local_vars, ssize_t id) { \n";
	oss << next_indent << "\treturn " << boost::apply_visitor(abort_visitor(a, u, syms.remote, local_symbol), e.expr) << ";\n";
	oss << next_indent << "}\n";

	oss << next_indent << "static logic::function_call logic_expression() {\n";
	oss << next_indent << "\treturn " << hyper::compiler::generate_logic_expression(e, a, u) << ";\n";
	oss << next_indent << "}\n";

	oss << indent << "};" << std::endl;
}

void genom_extension::output_import(std::ostream& oss, const functionDef& f, const typeList& tList) const
{
	// do nothing, we don't import to import genom extension to other module,
	// there is no reason they directly use it
}

void genom_extension::generate_expression_caller(std::ostream& oss, size_t counter) const
{
	std::string indent = times(2, "\t");
	oss << indent << "\tgenom_expression_" << counter << "::updater_type updater" << counter << ";\n";
	oss << indent << "\thyper::genom_model::compute_expression<genom_expression_" << counter ;
	oss << "> expression_exec" << counter << ";\n";
}


std::string genom_extension::generate_abortable_function(const universe& u,
		const ability& a, const task& t, const std::string& type_local_symbol,
		size_t counter, const std::string& target_var) const 
{
	std::ostringstream oss;
	std::string indent = times(3, "\t");
	std::string next_indent = "\t" + indent;

	oss << next_indent << "new hyper::model::abortable_function(\n";
	oss << next_indent << "boost::bind(&hyper::genom_model::compute_expression<genom_expression_" << counter << ">::async_eval<\n";
	oss << next_indent << "hyper::model::abortable_computation::cb_type,\n";
	oss << next_indent << "hyper::" << a.name() << "::ability,\n";
	oss << next_indent << type_local_symbol << " >,\n";
	oss << next_indent << "&expression_exec" << counter << ", _1, boost::ref(a), boost::cref(local_vars),\n";
	oss << next_indent << "boost::ref(" << target_var << ")),\n"; 
	oss << next_indent << "boost::bind(&hyper::genom_model::compute_expression<genom_expression_" << counter << ">::abort<\n";
	oss << next_indent << "hyper::" << a.name() << "::ability, " << type_local_symbol << ">,\n";
	oss << next_indent << "&expression_exec" << counter << ",boost::ref(a), boost::cref(local_vars)),\n";
	oss << next_indent << "boost::bind(&hyper::genom_model::compute_expression<genom_expression_" << counter << ">::error,\n";
	oss << next_indent << "&expression_exec" << counter << "))\n";

	counter++;

	return oss.str();
}

void genom_extension::recipe_additional_includes(std::ostream& oss, const std::string& ability_name) const
{
	oss << "#include <" << ability_name << "/opaque_types.hh>\n";
	oss << "#include <genom_model/compute_expression.hh>\n";
}
void
generate_cmakelist(std::ostream& oss, const std::string& name) 
{
	std::string NAME = hyper::compiler::upper(name);

	oss << "include(FindPkgConfig)\n";
	oss << "pkg_check_modules(" << NAME << " REQUIRED " << name << "-tclserv_client)\n";
	oss << "pkg_check_modules(POCOLIBS REQUIRED pocolibs)\n\n";

	oss << "include_directories(${" << NAME << "_INCLUDE_DIRS})\n";
	oss << "link_directories(${" << NAME << "_LIBRARY_DIRS})\n";
	oss << "include_directories(${POCOLIBS_INCLUDE_DIRS})\n";
	oss << "link_directories(${POCOLIBS_LIBRARY_DIRS})\n";

	oss << "set(USER_LIBS ${" << NAME << "_LIBRARIES} ${POCOLIBS_LIBRARIES})\n";
	oss << "set(USER_SRCS " << name << "_genom_interface.cc;";
	oss << "opaque_type_" << name << "Module.cc)";
}

void
generate_genom_interface_cc(std::ostream& oss, const std::string& name)
{
	std::string pattern = "#include \"@name@_genom_interface.hh\"\n"
						  "#include <genom_model/generate_module_interface.hh>\n"
						  "#define MODULE_NAME @name@\n\n"
						  "#include <boost/system/error_code.hpp>\n\n"
						  "ERROR_IMPLEMENTATION";

	oss << hyper::compiler::replace_by(pattern, "@name@", name);
}

void
generate_genom_interface_hh(std::ostream& oss, const std::string& name)
{
	std::string pattern = 
			"#ifndef @name@_GENOM_INTERFACE_HH_\n"
			"#define @name@_GENOM_INTERFACE_HH_\n\n"
			"#include <boost/asio.hpp>\n"
			"#include <tclserv_client/tclserv_client.h>\n"
			"#include <@name@/server/tclservClient/@name@TclservClientMsgLib.h>\n"
			"#include <@name@/server/tclservClient/@name@Error.h>\n\n"
			"#include <genom_model/genom_tclserv_client_module.hh>\n"
			"#include <genom_model/genom_process.hh>\n"
			"#include <genom_model/generate_module_interface.hh>\n"
			"#include <genom_model/poster.hh>\n\n"
			"#define MODULE_NAME @name@\n"
			"// XXX TODO COMPLETE\n"
			"#define MODULE_REQUESTS \\\n"
			"	    BOOST_PP_TUPLE_TO_LIST( \\\n"
			"				            0,              \\\n"
			"				            (               \\\n"
			"				            )                \\\n"
			"				    )\n\n"
			"// XXX TODO COMPLETE\n"
			"#define MODULE_POSTERS \\\n"
			"	    BOOST_PP_TUPLE_TO_LIST( \\\n"
			"				            0,              \\\n"
			"				            (               \\\n"
			"				            )               \\\n"
			"				    )\n\n"
			"	MODULE_DEFINITION\n\n"
			"#undef MODULE_NAME\n"
			"#undef MODULE_REQUESTS\n"
			"#undef MODULE_POSTERS\n"
			"#endif /* @name@_GENOM_INTERFACE_HH_ */ ";

	oss << hyper::compiler::replace_by(pattern, "@name@", name);
}

void
generate_opaque_module_hh(std::ostream& oss, const std::string& name)
{
	std::string pattern = 
		"#ifndef _@name@OPAQUE_TYPE_@name@Module_HH_\n"
		"#define _@name@OPAQUE_TYPE_@name@Module_HH_\n\n"
		"#include <boost/asio.hpp>\n"
		"#include <boost/function/function1.hpp>\n"
		"#include <boost/shared_ptr.hpp>\n\n"
		"namespace hyper {\n"
		"    namespace genom_model {\n"
		"        typedef boost::function<void (const boost::system::error_code& e)> cb_type;\n"
		"    }\n"
		"}\n\n"
		"namespace hyper {\n"
		"    namespace genom_model {\n"
		"        struct @name@;\n"
		"    }\n\n"
		"    namespace @name@ {\n"
		"            struct @name@Module\n"
		"            {\n"
		"                boost::shared_ptr< genom_model::@name@ > impl;\n\n"
		"                @name@Module();\n"
		"                @name@Module(boost::asio::io_service& io_s,\n"
		"                          const std::string& addr,\n"
		"                          int port,\n"
		"                          hyper::genom_model::cb_type cb);\n"
		"            };\n\n"
		"	}\n"
		"}\n\n"
		"#endif /* _@name@OPAQUE_TYPE_@name@Module_HH_ */ ";

	oss << hyper::compiler::replace_by(pattern, "@name@", name);
}

void
generate_opaque_module_cc(std::ostream& oss, const std::string& name)
{
	std::string pattern = 
		"#include <@name@_genom_interface.hh>\n"
		"#include <opaque_type_@name@Module.hh>\n\n"
		"namespace hyper {\n"
		"    namespace @name@ {\n\n"
		"           @name@Module::@name@Module() {}\n\n"
		"           @name@Module::@name@Module(boost::asio::io_service& io_s,\n"
		"                                 const std::string& addr,\n"
		"                                 int port,\n"
		"                                 genom_model::@name@::cb_type cb)\n"
		"            {\n"
		"                std::ostringstream oss;\n"
		"                oss << port;\n"
		"                impl.reset(new genom_model::@name@(io_s, addr, oss.str(), cb));\n"
		"            }\n"
		"    }\n"
		"}";

	oss << hyper::compiler::replace_by(pattern, "@name@", name);
}

void 
genom_extension::generate_additional_files(const std::string& path,
										   const universe & u,
										   const std::string& name) const 
{
	{
		std::string fileName = path + "/CMakeLists.txt";
		std::ofstream oss(fileName.c_str());
		generate_cmakelist(oss, name);
	}
	{
		std::string fileName = path + "/" + name + "_genom_interface.hh";
		std::ofstream oss(fileName.c_str());
		generate_genom_interface_hh(oss, name);
	}

	{
		std::string fileName = path + "/" + name + "_genom_interface.cc";
		std::ofstream oss(fileName.c_str());
		generate_genom_interface_cc(oss, name);
	}

	{
		std::string fileName = path + "/opaque_type_" + name + "Module.hh";
		std::ofstream oss(fileName.c_str());
		generate_opaque_module_hh(oss, name);
	}

	{
		std::string fileName = path + "/opaque_type_" + name + "Module.cc";
		std::ofstream oss(fileName.c_str());
		generate_opaque_module_cc(oss, name);
	}
}

